package dev.fitko.fitconnect.core;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.crypto.RSASSASigner;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventClaimFields;
import dev.fitko.fitconnect.api.domain.model.event.EventHeaderFields;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.event.SubmissionState.ACCEPTED;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class SubmissionSenderTest {

    private SubmissionService submissionServiceMock;
    private EventLogService eventLogServiceMock;
    private CryptoService cryptoServiceMock;
    private ValidationService validationServiceMock;
    private KeyService keyServiceMock;
    private Sender underTest;

    @BeforeEach
    public void setUp() {
        submissionServiceMock = mock(SubmissionService.class);
        eventLogServiceMock = mock(EventLogService.class);
        cryptoServiceMock = mock(CryptoService.class);
        validationServiceMock = mock(ValidationService.class);
        keyServiceMock = mock(KeyService.class);

        underTest = new SubmissionSender(
                submissionServiceMock,
                eventLogServiceMock,
                cryptoServiceMock,
                validationServiceMock,
                keyServiceMock
        );
    }

    @Test
    void validatePublicKey() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(validationServiceMock.validateEncryptionPublicKey(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(publicKey);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void encryptBytes() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(cryptoServiceMock.encryptBytes(any(), any())).thenReturn("encryptedString");

        // When
        final String encryptedData = underTest.encryptBytes(publicKey, "data".getBytes());

        // Then
        assertNotNull(encryptedData);
        assertThat(encryptedData, is("encryptedString"));
    }

    @Test
    void createHash() {

        // Given
        when(cryptoServiceMock.hashBytes(any())).thenReturn("hashedData");

        // When
        final String hashedData = underTest.createHash("data".getBytes());

        // Then
        assertNotNull(hashedData);
        assertThat(hashedData, is("hashedData"));
    }

    @Test
    void createSubmission() {

        // Given
        final var createSubmission = CreateSubmission.builder()
                .caseId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .announcedAttachments(Collections.emptyList())
                .serviceType(new ServiceType())
                .build();

        final var expectedSubmission = new SubmissionForPickup();
        expectedSubmission.setSubmissionId(UUID.randomUUID());
        expectedSubmission.setDestinationId(createSubmission.getDestinationId());
        expectedSubmission.setCaseId(createSubmission.getCaseId());

        when(submissionServiceMock.announceSubmission(any())).thenReturn(expectedSubmission);

        // When
        final SubmissionForPickup submissionForPickup = underTest.createSubmission(createSubmission);

        // Then
        assertThat(submissionForPickup, equalTo(expectedSubmission));
    }

    @Test
    void uploadAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "encryptedAttachmentBytes";

        doNothing().when(submissionServiceMock).uploadAttachment(any(), any(), any());

        // When
        underTest.uploadAttachment(submissionId, attachmentId, encryptedAttachment);

        // Then
        verify(submissionServiceMock, times(1)).uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Test
    void getEncryptionKeyForDestination() throws ParseException, IOException {

        // Given
        final var submissionId = UUID.randomUUID();

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final RSAKey expectedRSAKey = RSAKey.parse(publicKeyJson);

        when(keyServiceMock.getPublicEncryptionKey(any())).thenReturn(expectedRSAKey);

        // When
        final RSAKey key = underTest.getEncryptionKeyForDestination(submissionId);

        // Then
        assertThat(key, equalTo(expectedRSAKey));
    }

    @Test
    void getEncryptionKeyForDestinationFailed() {

        // Given
        when(keyServiceMock.getPublicEncryptionKey(any())).thenThrow(RestApiException.class);

        // Then
        assertThrows(
                RestApiException.class,
                () -> underTest.getEncryptionKeyForDestination(UUID.randomUUID()),
                "Expected getEncryptionKeyForDestination() to throw, but nothing was thrown"
        );
    }

    @Test
    void encryptObject() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(cryptoServiceMock.encryptObject(any(), any())).thenReturn("encryptedObject");

        final Metadata metadata = new Metadata();

        // When
        final String encryptedMetadata = underTest.encryptObject(publicKey, metadata);

        // Then
        assertThat(encryptedMetadata, equalTo("encryptedObject"));
    }

    @Test
    void encryptingObjectFailed() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        // When
        when(cryptoServiceMock.encryptObject(any(), any())).thenThrow(EncryptionException.class);

        // Then
        assertThrows(
                EncryptionException.class,
                () -> underTest.encryptObject(publicKey, new Metadata()),
                "Expected encryptObject() to throw, but nothing was thrown"
        );
    }

    @Test
    void sendSubmission() {

        // Given
        final SubmitSubmission submitSubmission = new SubmitSubmission();
        submitSubmission.setSubmissionId(UUID.randomUUID());
        submitSubmission.setEncryptedData("encryptedData");
        submitSubmission.setEncryptedMetadata("encryptedMetadata");

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submitSubmission.getSubmissionId());
        expectedSubmission.setEncryptedData(submitSubmission.getEncryptedData());
        expectedSubmission.setEncryptedMetadata(submitSubmission.getEncryptedMetadata());

        when(submissionServiceMock.sendSubmission(any())).thenReturn(expectedSubmission);

        // When
        underTest.sendSubmission(submitSubmission);

        // Then
        verify(submissionServiceMock, times(1)).sendSubmission(submitSubmission);
    }

    @Test
    void getDestination() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedDestination = new Destination();
        expectedDestination.setDestinationId(destinationId);

        when(submissionServiceMock.getDestination(any())).thenReturn(expectedDestination);

        // When
        final Destination destination = underTest.getDestination(destinationId);

        // Then
        assertThat(destination, equalTo(expectedDestination));
    }

    @Test
    void testValidMetadata() {

        // Given
        final var metadata = new Metadata();

        when(validationServiceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validatedMetadata = underTest.validateMetadata(metadata);

        // Then
        assertTrue(validatedMetadata.isValid());
    }

    @Test
    void testInvalidMetadata() {

        // Given
        final var invalidMetadata = new Metadata();
        final ValidationResult expectedResult = ValidationResult.error(new ValidationException("invalid metadata"));
        when(validationServiceMock.validateMetadataSchema(any())).thenReturn(expectedResult);

        // When
        final ValidationResult validatedMetadata = underTest.validateMetadata(invalidMetadata);

        // Then
        assertTrue(validatedMetadata.hasError());
        assertThat(validatedMetadata.getError().getMessage(), equalTo("invalid metadata"));
    }

    @Test
    void testValidJsonFormatValidation() {

        // Given
        when(validationServiceMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateJsonFormat("{}");

        //Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidJsonFormatValidation() {

        // Given
        when(validationServiceMock.validateJsonFormat(any())).thenReturn(ValidationResult.error(new ValidationException("Invalid json")));

        // When
        final ValidationResult validationResult = underTest.validateJsonFormat("$%&/()");

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid json"));
    }

    @Test
    void testValidXmlFormatValidation() {

        // Given
        when(validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateXmlFormat("<xml>test</xml>");

        //Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidXmlFormatValidation() {

        // Given
        when(validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.error(new ValidationException("Invalid xml")));

        // When
        final ValidationResult validationResult = underTest.validateXmlFormat("$%&/(");

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid xml"));
    }

    @Test
    void testReceiveValidEventLog() {

        // Given
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();

        final EventLogEntry eventLogEntry = EventLogEntry.builder()
                .caseId(caseId)
                .build();

        when(eventLogServiceMock.getEventLog(any(), any())).thenReturn(List.of(eventLogEntry));

        // When
        final List<EventLogEntry> eventLog = underTest.getEventLog(destinationId, caseId);

        // Then
        assertFalse(eventLog.isEmpty());
        assertThat(eventLog.get(0).getCaseId(), is(caseId));
    }

    @Test
    void testReceiveEventLogWithInvalidValidation() {

        // Given
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();

        when(eventLogServiceMock.getEventLog(any(), any())).thenThrow(new EventLogException("Invalid log"));

        // Then
        assertThrows(EventLogException.class, () -> underTest.getEventLog(destinationId, caseId));
    }

    @Test
    void testReceiveLatestStatusForSubmission() {

        // Given
        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final SubmissionStatus expectedStatus = new SubmissionStatus(ACCEPTED, Collections.emptyList());

        when(eventLogServiceMock.getLastedEvent(any(), any(), any(), any())).thenReturn(expectedStatus);

        // When
        final SubmissionStatus status = underTest.getLastedEvent(destinationId, caseId, submissionId, new AuthenticationTags());

        // Then
        assertThat(status.getStatus(), is(ACCEPTED));
    }

    private SignedJWT getSignedJWT(final UUID destinationId, final UUID caseId, final UUID submissionId, final Event event) throws JOSEException, IOException, ParseException {
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);
        final JWTClaimsSet claims = getClaimsForEvent(caseId, destinationId, submissionId, event);
        final SignedJWT signedJWT = new SignedJWT(header, claims);
        signedJWT.sign(new RSASSASigner(rsaKey));
        return signedJWT;
    }

    private JWTClaimsSet getClaimsForEvent(final UUID caseId, final UUID destinationId, final UUID submissionId, final Event event) {

        final AuthenticationTags authTags = new AuthenticationTags();
        authTags.setData("dataAuthTag");
        authTags.setMetadata("metadataAuthTag");
        authTags.setAttachments(Map.of(UUID.randomUUID(), "attachmentAuthTag"));

        final Map<String, Object> tags = Map.of("data", authTags.getData(), "metadata", authTags.getMetadata(), "attachments", authTags.getAttachments());
        final Map<String, Object> payload = Map.of("authenticationTags", tags, "problems", Collections.emptyList());

        final Map<String, Map<String, Object>> events = Map.of(event.getSchemaUri(), payload);

        return new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
    }

    private JWSHeader getDefaultJwsHeader(final RSAKey rsaKey) throws ParseException {
        return JWSHeader.parse(Map.of(
                EventHeaderFields.TYPE, "secevent+jwt",
                EventHeaderFields.KEY_ID, rsaKey.getKeyID(),
                EventHeaderFields.ALGORITHM, "PS512"
        ));
    }

    private RSAKey getRsaKey() throws IOException, ParseException {
        return RSAKey.parse(getResourceAsString("/private_test_verify_key.json"));
    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(SubmissionSenderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}